//===- SPIRVImageInterfaces.cpp ---------------------------------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#include "mlir/IR/OpDefinition.h"

#include "mlir/Dialect/SPIRV/Interfaces/SPIRVImageInterfaces.h"

using namespace mlir;

#include "mlir/Dialect/SPIRV/Interfaces/SPIRVImageInterfaces.cpp.inc"
