//===--- NVVMRequiresSMTraits.cpp - NVVM Requires SM Traits -----*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
//
// This file defines op traits for the NVVM Dialect in MLIR
//
//===----------------------------------------------------------------------===//

#include "mlir/Dialect/LLVMIR/NVVMRequiresSMTraits.h"

#include "mlir/Dialect/LLVMIR/NVVMRequiresSMTraits.cpp.inc"
