// This is a truncated version of lldb-defines.h used to test the script
// that fixes up its versioning info.

// The script needs to uncomment these lines and populate the info for versioning.
// #define LLDB_VERSION
// #define LLDB_REVISION
// #define LLDB_VERSION_STRING
