#import <Foundation/Foundation.h>

int main()
{
    @autoreleasepool
    {
        NSString *string = @"Hello";
        NSArray *array = @[ @1, @2, @3 ];

        NSLog(@"Stop here"); // Set breakpoint 0 here.
    }
}
