//===- IndexDialect.h - Index dialect declaration -----------------*- C++-*-==//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef MLIR_DIALECT_INDEX_IR_INDEXDIALECT_H
#define MLIR_DIALECT_INDEX_IR_INDEXDIALECT_H

#include "mlir/IR/Dialect.h"

//===----------------------------------------------------------------------===//
// ODS-Generated Declarations
//===----------------------------------------------------------------------===//

#include "mlir/Dialect/Index/IR/IndexOpsDialect.h.inc"

#endif // MLIR_DIALECT_INDEX_IR_INDEXDIALECT_H
