//===- SMTDialect.h - SMT dialect definition --------------------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef MLIR_DIALECT_SMT_IR_SMTDIALECT_H
#define MLIR_DIALECT_SMT_IR_SMTDIALECT_H

#include "mlir/IR/BuiltinOps.h"
#include "mlir/IR/Dialect.h"
#include "mlir/Support/LLVM.h"

// Pull in the dialect definition.
#include "mlir/Dialect/SMT/IR/SMTDialect.h.inc"
#include "mlir/Dialect/SMT/IR/SMTEnums.h.inc"

#endif // MLIR_DIALECT_SMT_IR_SMTDIALECT_H
