//===- StateStack.cpp - Utility for storing a stack of state --------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#include "mlir/Support/StateStack.h"

namespace mlir {

void StateStackFrame::anchor() {}

} // namespace mlir
