// This ia a basic header file used to check that the server-side emitter
// for rpc-gen emits an expected set of includes in a generated source file.
#ifndef LLDB_API_SBRPC_CHECKBASICINCLUDE_H
#define LLDB_API_SBRPC_CHECKBASICINCLUDE_H

#endif // LLDB_API_SBRPC_CHECKBASICINCLUDE_H
