//===- PtrEnums.h - `ptr` dialect enums -------------------------*- C++ -*-===//
//
// This file is licensed under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
//
// This file declares the `ptr` dialect enums.
//
//===----------------------------------------------------------------------===//

#ifndef MLIR_DIALECT_PTR_IR_PTRENUMS_H
#define MLIR_DIALECT_PTR_IR_PTRENUMS_H

#include "mlir/IR/BuiltinAttributeInterfaces.h"
#include "mlir/IR/OpImplementation.h"

#include "mlir/Dialect/Ptr/IR/PtrOpsEnums.h.inc"

#endif // MLIR_DIALECT_PTR_IR_PTRENUMS_H
