#define T unsigned short
#define T_CSTR "unsigned short"
#define T_VALUE_1 11001
#define T_VALUE_2 22002
#define T_VALUE_3 0
#define T_VALUE_4 65535
#define T_PRINTF_FORMAT "%hu"

#include "basic_type.cpp"
