//===----------------------------------------------------------------------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//
///
/// \file
/// This file contains the definition of type aliases for extended floating
/// -point types.
///
//===----------------------------------------------------------------------===//

#ifndef MATHTEST_TYPEEXTRAS_HPP
#define MATHTEST_TYPEEXTRAS_HPP

namespace mathtest {

using float16 = _Float16;
} // namespace mathtest

#endif // MATHTEST_TYPEEXTRAS_HPP
