//===- TransformAttrs.cpp - Transform Dialect Attribute Definitions -------===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#include "mlir/Dialect/Transform/IR/TransformAttrs.h"
#include "mlir/IR/BuiltinTypes.h"

#include "mlir/Dialect/Transform/IR/TransformDialectEnums.cpp.inc"
