// We need to import any std module. It doesn't matter which one.
#include <iostream>

int main(int argc, char **argv) {
  std::cout << "Test" << std::endl;
  return 0; // Set break point at this line.
}
