/*===----- llvm/Config/Targets.h - llvm targets configuration -----*- C -*-===*/
/*                                                                            */
/* Part of the LLVM Project, under the Apache License v2.0 with LLVM          */
/* Exceptions.                                                                */
/* See https://llvm.org/LICENSE.txt for license information.                  */
/* SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception                    */
/*                                                                            */
/*===----------------------------------------------------------------------===*/

// Bazel builds provides target #defines through @llvm-project//llvm:config.
// This file is therefore empty, but it exists so it can be #included.
