#!/usr/bin/env bash

echo "<?xml version=\"1.0\" encoding=\"UTF-8\"?>"
echo "<!DOCTYPE plist PUBLIC \"-//Apple//DTD PLIST 1.0//ENhttp://www.apple.com/DTDs/PropertyList-1.0.dtd\">"
echo "<plist version=\"1.0\">"
echo "<dict>"
echo "    <key>41945CA4-5D9D-3CDE-82B4-37E4C09750B5</key>"
echo "    <dict>"
echo "        <key>DBGArchitecture</key>"
echo "        <string>x86_64</string>"
echo "        <key>DBGError</key>"
echo "        <string>UUID information was not found</string>"
echo "    </dict>"
echo "</dict>"
echo "</plist>"
