#define T double
#define T_CSTR "double"
#define T_VALUE_1 1100.125
#define T_VALUE_2 2200.250
#define T_VALUE_3 33.00
#define T_VALUE_4 44.00
#define T_PRINTF_FORMAT "%lg"

#include "basic_type.cpp"
