#include <vector>

typedef std::vector<long> vector_long;
int main(int argc, char **argv) {
  std::vector<int> a = {3, 1, 2};
  vector_long b = {3, 1, 2};
  return 0; // Set break point at this line.
}
