int main() {
  // Disable clang-format as it gets confused by the keyword identifiers.
  // clang-format off
  int alignas = 1;
  int alignof = 1;
  int and = 1;
  int and_eq = 1;
  int atomic_cancel = 1;
  int atomic_commit = 1;
  int atomic_noexcept = 1;
  int bitand = 1;
  int bitor = 1;
  int bool = 1;
  int catch = 1;
  int char8_t = 1;
  int char16_t = 1;
  int char32_t = 1;
  int class = 1;
  int compl = 1;
  int concept = 1;
  int consteval = 1;
  int constexpr = 1;
  int constinit = 1;
  int const_cast = 1;
  int co_await = 1;
  int co_return = 1;
  int co_yield = 1;
  int decltype = 1;
  int delete = 1;
  int dynamic_cast = 1;
  int explicit = 1;
  int export = 1;
  int false = 1;
  int friend = 1;
  int mutable = 1;
  int namespace = 1;
  int new = 1;
  int noexcept = 1;
  int not = 1;
  int not_eq = 1;
  int operator= 1;
  int or = 1;
  int or_eq = 1;
  int private = 1;
  int protected = 1;
  int public = 1;
  int reflexpr = 1;
  int reinterpret_cast = 1;
  int requires = 1;
  int static_assert = 1;
  int static_cast = 1;
  int synchronized = 1;
  int template = 1;
  int this = 1;
  int thread_local = 1;
  int throw = 1;
  int true = 1;
  int try = 1;
  int typeid = 1;
  int typename = 1;
  int using = 1;
  int virtual = 1;
  int wchar_t = 1;
  int xor = 1;
  int xor_eq = 1;
  // clang-format on
  return 0; // break here
}
